Require Import FunctionalExtensionality List Omega.

Require Import PropXTac.

Set Implicit Arguments.


Module Type S.
  Parameters addr byte : Type.
  Parameter addr_eq : forall x y : addr, {x = y} + {x <> y}.

  Parameter state : Type.
  Parameter Mem : state -> addr -> byte.
End S.

Module Make (M : S).
  Definition smem := M.addr -> option M.byte.

  Definition disjoint (m1 m2 : smem) :=
    forall a v, m1 a = Some v -> m2 a = None.

  Theorem disjoint_symm : forall m1 m2, disjoint m1 m2 -> disjoint m2 m1.
    unfold disjoint; intros.
    specialize (H a).
    destruct (m1 a); auto.
    erewrite H in H0; eauto; discriminate.
  Qed.

  Definition hprop := smem -> Prop.

  Definition pure (P : Prop) : hprop := fun sm => P /\ forall a, sm a = None.

  Definition split (sm sm1 sm2 : smem) : Prop :=
    disjoint sm1 sm2
    /\ forall a, sm a = match sm1 a with
                          | None => sm2 a
                          | v => v
                        end.

  Definition star (p1 p2 : hprop) : hprop := fun sm =>
    exists sm1, exists sm2, split sm sm1 sm2 /\ p1 sm1 /\ p2 sm2.

  Definition exis T (p : T -> hprop) : hprop := fun sm =>
    exists v, p v sm.

  Definition ptsTo (a : M.addr) (v : M.byte) : hprop := fun sm =>
    sm a = Some v /\ forall a', a' <> a -> sm a' = None.

  Inductive sprop : Type :=
  | Pure : Prop -> sprop
  | Impure : hprop -> sprop
  | Star : sprop -> sprop -> sprop
  | Exis : forall T, (T -> sprop) -> sprop.

  (* You'll probably want a notation like this, with an appropriate inner scope:
     Notation "[< p >]" := (Pure p%nat) : Sep_scope.
     This isn't hardcoded here because the scope depends on factors like which type you use for addresses. *)

  Infix "*" := Star : Sep_scope.

  Delimit Scope Sep_scope with Sep.
  Bind Scope Sep_scope with sprop.

  Notation "'Ex' x , P" := (Exis (fun x => P%Sep)) (x ident, at level 89) : Sep_scope.
  Notation "'Ex' x : A , P" := (Exis (fun x : A => P%Sep)) (x ident, at level 89) : Sep_scope.
  Notation "a ==> v" := (Impure (ptsTo a v)) (no associativity, at level 39) : Sep_scope.
  Notation "![ p ]" := (Impure p) : Sep_scope.

  Definition emp : sprop := Pure True.

  Definition stars' := fold_right star (pure True).

  Module Type STARS_DEF.
    Parameter stars : list hprop -> hprop.
    Axiom stars_def : stars = stars'.
  End STARS_DEF.

  Module StarsDef : STARS_DEF.
    Definition stars := stars'.

    Theorem stars_def : stars = stars'.
      reflexivity.
    Qed.
  End StarsDef.

  Import StarsDef.
  Export StarsDef.

  Fixpoint sep' (p : sprop) (k : Prop -> list hprop -> Prop) : Prop :=
    match p with
      | Pure P => k P nil
      | Impure p => k True (p :: nil)
      | Star p1 p2 => sep' p1 (fun P1 ps1 =>
        sep' p2 (fun P2 ps2 =>
          k (P1 /\ P2) (ps1 ++ ps2)))
      | Exis _ p1 => exists x, sep' (p1 x) k
    end.

  Definition sep (p : sprop) sm := sep' p (fun P ps => P /\ stars ps sm).

  (* You'll probably want a machine-dependent notation like this one:
     Notation "![ p ]" := (fun st => [< sep p%Sep (fun a => Some (Mem st a)) >]) : PropX_scope. *)

  Notation "!{ p }" := (Impure (sep p)) : Sep_scope.
  
  Definition Himp (p p' : sprop) := forall sm, sep p sm -> sep p' sm.

  Notation "p1 ===> p2" := (Himp p1%Sep p2%Sep) (no associativity, at level 90).

  Ltac splitmaster := unfold split, disjoint, star; intuition;
    repeat match goal with
             | [ a : M.addr, H : forall a : M.addr, _ |- _ ] => specialize (H a)
             | [ a : M.addr |- _ ] =>
               match goal with
                 | [ |- ?sm a = _ ] => destruct (sm a); try congruence
               end
             | [ H : context[match ?E with None => _ | Some _ => _ end] |- _ ] => destruct E; try congruence
             | [ |- context[match ?E with None => _ | Some _ => _ end] ] => destruct E; try congruence
             | [ H : _, H' : _ = _ |- _ ] => erewrite H in H'; congruence || reflexivity
             | [ H : _ |- _ ] => erewrite H; congruence || reflexivity
             | [ H : _ |- _ ] => specialize (H _ (refl_equal _)); congruence
           end.

  Theorem split_sym : forall sm sm1 sm2,
    split sm sm1 sm2
    -> split sm sm2 sm1.
    splitmaster.
  Qed.

  Hint Immediate split_sym.

  Theorem ptsTo_contra : forall a v (ps2 ps1 : list hprop) sm,
    stars' (ps1 ++ ptsTo a v :: ps2) sm
    -> sm a <> None.
    induction ps1; simpl; intuition.

    red in H; firstorder.
    splitmaster.

    red in H; firstorder.
    specialize (H2 a).
    rewrite H0 in H2.
    red in H.
    specialize (H a).
    destruct (x a); try discriminate.
    eauto.
  Qed.

  Theorem read : forall a a' v (ps2 ps1 : list hprop) sm,
    stars (ps1 ++ ptsTo a v :: ps2) sm
    -> a' = a
    -> sm a' = Some v.
    rewrite stars_def; induction ps1; simpl; intuition; subst.

    red in H; firstorder.
    splitmaster.
    
    red in H; firstorder.
    specialize (IHps1 _ H2 (refl_equal _)).
    specialize (H a).
    splitmaster.
  Qed.

  Theorem ptrsNeq : forall ps1 ps2 ps3 a1 a2 a1' a2' v1 v2 sm,
    stars (ps1 ++ ptsTo a1' v1 :: ps2 ++ ptsTo a2' v2 :: ps3) sm
    -> a1' = a1
    -> a2' = a2
    -> a1 <> a2.
    rewrite stars_def; induction ps1; simpl; intuition; subst; eauto.

    red in H; firstorder.
    eapply ptsTo_contra.
    eauto.
    specialize (H a2); rewrite H0 in H; eauto.

    red in H; firstorder.
    eauto.
  Qed.

  Theorem overwrite : forall a a' v v' ps2 ps1 sm,
    stars (ps1 ++ ptsTo a v :: ps2) sm
    -> a' = a
    -> stars (ps1 ++ ptsTo a v' :: ps2)
      (fun a'' => if M.addr_eq a'' a' then Some v' else sm a'').
    rewrite stars_def; induction ps1; simpl; intuition; subst.

    unfold star in *; firstorder.
    exists (fun a'' => if M.addr_eq a'' a then Some v' else None).
    exists x0; intuition.
    
    red; intuition.
    red; intros.
    destruct (M.addr_eq a0 a); subst; try discriminate.
    specialize (H a).
    eauto.
    
    destruct (M.addr_eq a0 a); subst; try congruence.
    rewrite H1.
    erewrite H3; auto.

    red; intuition.
    destruct (M.addr_eq a a); tauto.
    destruct (M.addr_eq a' a); tauto.

    red in H; firstorder.
    exists x.
    exists (fun a'' => if M.addr_eq a'' a then Some v' else x0 a''); intuition.
    red; intuition.
    red; intuition.
    destruct (M.addr_eq a1 a); subst.
    specialize (ptsTo_contra _ _ _ _ _ H2).
    specialize (H _ _ H3).
    tauto.

    apply (H _ _ H3).

    destruct (M.addr_eq a1 a); subst.

    specialize (ptsTo_contra _ _ _ _ _ H2).
    specialize (H a).
    destruct (x a); intuition.
    elimtype False; eauto.
    auto.
  Qed.

  Definition himp (ps ps' : list hprop) := forall sm, stars ps sm -> stars ps' sm.

  Infix "--->" := himp (no associativity, at level 90).

  Theorem stars_final : forall ls ls' sm sm',
    stars ls sm
    -> (forall a, sm a = sm' a)
    -> ls ---> ls'
    -> stars ls' sm'.
    intros; replace sm' with sm; auto.
    extensionality x; auto.
  Qed.

  Theorem himp_refl : forall ps, ps ---> ps.
    red; auto.
  Qed.

  Hint Unfold pure.

  Lemma split_right : forall sm x x0,
    split sm x x0
      -> (forall a, x0 a = None)
      -> x = sm.
    unfold split, disjoint; intuition.
    extensionality a.
    specialize (H0 a); specialize (H1 a); specialize (H2 a).
    rewrite H0 in H2.
    destruct (x a); auto.
  Qed.

  Theorem sep'_weaken : forall p k, sep' p k
    -> forall k' : _ -> _ -> Prop, (forall P ps, k P ps -> k' P ps)
      -> sep' p k'.
    induction p; simpl; firstorder.

    eapply IHp1; eauto.
    simpl; eauto.
  Qed.

  Lemma frame_finish : forall ps,
    ps ---> sep (fold_right Star (Pure True) (map Impure ps)) :: nil.
    unfold himp; induction ps; simpl; intuition; rewrite stars_def in *; simpl in *.

    unfold sep; simpl.
    rewrite stars_def; simpl.
    exists sm; exists (fun _ => None); intuition.
    red; intuition.
    red; intuition.
    destruct (sm a); tauto.
    red; auto.

    red in H; firstorder.
    specialize (IHps _ H2).
    exists sm; exists (fun _ => None); intuition.
    splitmaster.
    red in IHps; firstorder.
    generalize (split_right (conj H3 H5) H7); intro; subst.
    unfold sep in *; simpl in *.
    eapply sep'_weaken.
    eauto.
    simpl; intuition.
    rewrite stars_def in *; simpl.
    exists x; exists x0; intuition.
    red; tauto.
  Qed.

  Theorem himp_cancel_front : forall p ps ps',
    ps ---> ps'
    -> p :: ps ---> p :: ps'.
    unfold himp; rewrite stars_def; simpl; intuition.

    unfold star in *; firstorder.
  Qed.

  Lemma himp_pullout' : forall p p' ps,
    p :: p' :: ps ---> p' :: p :: ps.
    unfold himp; rewrite stars_def; simpl; firstorder.

    exists x1; exists (fun a => match x a with
                                  | None => x2 a
                                  | v => v
                                end); intuition.
    unfold split, disjoint in *; intuition.

    splitmaster.
    splitmaster.

    exists x; exists x2; intuition.
    unfold split, disjoint in *; intuition.
    specialize (H _ _ H6).
    specialize (H4 a).
    rewrite H in H4.
    specialize (H2 a).
    destruct (x1 a); congruence.
  Qed.

  Theorem himp_trans : forall p1 p2 p3,
    p1 ---> p2
    -> p2 ---> p3
    -> p1 ---> p3.
    unfold himp; eauto.
  Qed.

  Lemma himp_pullout1 : forall ps2 p ps1,
    ps1 ++ p :: ps2 ---> p :: ps1 ++ ps2.
    induction ps1; simpl; intuition.

    apply himp_refl.

    apply himp_trans with (a :: p :: ps1 ++ ps2).

    apply himp_cancel_front; eauto.

    apply himp_pullout'; eauto.
  Qed.

  Lemma himp_pullout2 : forall ps2 p ps1,
    p :: ps1 ++ ps2 ---> ps1 ++ p :: ps2.
    induction ps1; simpl; intuition.

    apply himp_refl.

    apply himp_trans with (a :: p :: ps1 ++ ps2).

    apply himp_pullout'; eauto.

    apply himp_cancel_front; eauto.
  Qed.

  Theorem himp_cancel : forall p ps2 ps2' ps1 ps1',
    ps1 ++ ps2 ---> ps1' ++ ps2'
    -> ps1 ++ p :: ps2 ---> ps1' ++ p :: ps2'.
    intros.
    apply himp_trans with (p :: ps1 ++ ps2).

    apply himp_pullout1; eauto.

    apply himp_trans with (p :: ps1' ++ ps2').

    apply himp_cancel_front; eauto.

    apply himp_pullout2; eauto.
  Qed.

  Theorem himp_cancel_eq : forall ps1 ps2 ps1' ps2' p p',
    p' = p
    -> ps1 ++ ps2 ---> ps1' ++ ps2'
    -> ps1 ++ p :: ps2 ---> ps1' ++ p' :: ps2'.
    intros; subst; apply himp_cancel; auto.
  Qed.

  Theorem stars_elim_Pure : forall P ps2 ps1 sm,
    stars (ps1 ++ sep (Pure P) :: ps2) sm
    -> P /\ stars (ps1 ++ ps2) sm.
    induction ps1; simpl; firstorder; rewrite stars_def in *; firstorder.

    rewrite stars_def in H3; simpl in H3.
    red in H3; intuition.

    replace sm with x0; auto.
    extensionality a.
    rewrite H1.
    rewrite H5.
    reflexivity.
  Qed.

  Lemma stars'_sep_fwd : forall ps2 sm,
    stars' ps2 sm
    -> sep (fold_right Star (Pure True) (map Impure ps2)) sm.
    induction ps2; simpl; firstorder.

    rewrite stars_def; firstorder.

    unfold sep in *; simpl in *.
    specialize (IHps2 _ H2).
    eapply sep'_weaken; eauto.
    simpl; intuition.
    rewrite stars_def in *; firstorder.
  Qed.

  Lemma split_shuffle : forall sm x x0 x1 x2,
    split sm x x0
    -> split x x1 x2
    -> split (fun a => match x1 a with
                         | None => sm a
                         | _ => None
                       end) x2 x0.
    splitmaster.
  Qed.

  Theorem desep_fwd : forall p ps2 ps1 sm, stars (ps1 ++ sep p :: ps2) sm
    -> sep (fold_right Star (Pure True) (map Impure ps1 ++ p :: map Impure ps2)) sm.
    rewrite stars_def; induction ps1; simpl; firstorder.

    unfold sep; simpl.
    eapply sep'_weaken.
    eauto.
    simpl; intuition.
    rewrite stars_def in *.
    clear H0.
    generalize dependent x0.
    generalize dependent x.
    generalize dependent sm.
    generalize dependent ps2.
    induction ps; simpl; firstorder.

    apply stars'_sep_fwd in H2.
    eapply sep'_weaken.
    eassumption.
    simpl; intuition.
    rewrite stars_def in *.
    replace sm with x0; eauto.
    extensionality a.
    rewrite H1; rewrite H3; reflexivity.

    specialize (IHps _ _ _ H6 _ (proj1 (split_shuffle (conj H H1) (conj H0 H5))) (proj2 (split_shuffle (conj H H1) (conj H0 H5))) H2).
    eapply sep'_weaken; [ eauto | ].
    simpl; intuition.

    exists x1.
    exists (fun a : M.addr =>
      match x1 a with
        | Some _ => None
        | None => sm a
      end).
    intuition.
    splitmaster.

    unfold sep in *; simpl.
    specialize (IHps1 _ H2).
    eapply sep'_weaken; [ eauto | ].
    simpl; intuition.
    rewrite stars_def in *; firstorder.
  Qed.

  Theorem unfold_sep : forall p p' ps2 ps1 sm,
    stars (ps1 ++ sep p :: ps2) sm
    -> p ===> p'
    -> stars (ps1 ++ sep p' :: ps2) sm.
    unfold Himp; rewrite stars_def; induction ps1; simpl; firstorder.
  Qed.

  Theorem unfold_sep_fwd : forall p p' ps2 ps1 sm,
    stars (ps1 ++ sep p :: ps2) sm
    -> p ===> p'
    -> sep (fold_right Star (Pure True)
      (map Impure ps1 ++ p' :: map Impure ps2)) sm.
    intros; apply desep_fwd; eapply unfold_sep; eauto.
  Qed.

  Lemma split_assoc : forall sm x x0 x1 x2,
    split sm x x0
      -> split x0 x1 x2
        -> split sm (fun a => match x a with
                                | None => x1 a
                                | v => v
                              end) x2.
    splitmaster.
  Qed.

  Lemma stars'_app_fwd : forall ps2 ps1 sm,
    stars' (ps1 ++ ps2) sm
    -> exists sm1, exists sm2, split sm sm1 sm2
      /\ stars' ps1 sm1
      /\ stars' ps2 sm2.
    induction ps1; simpl; firstorder.

    exists (fun _ => None); exists sm; intuition.
    splitmaster.

    specialize (IHps1 _ H2); firstorder.
    exists (fun a => match x a with
                       | None => x1 a
                       | v => v
                     end); exists x2.
    split.
    eapply split_assoc; eauto.
    red; eauto.
    red; intuition.
    intuition.
    exists x; exists x1; intuition eauto.
    red; intuition.
    unfold disjoint in *; intuition.
    specialize (H3 a0).
    specialize (H a0).
    specialize (H5 a0).
    destruct (x1 a0); auto.
    specialize (H1 a0).
    rewrite H7 in H1.
    specialize (H _ H7).
    congruence.
  Qed.

  Lemma stars'_app_bwd : forall ps2 sm2 ps1 sm sm1,
    split sm sm1 sm2
    -> stars' ps1 sm1
    -> stars' ps2 sm2
    -> stars' (ps1 ++ ps2) sm.
    induction ps1; simpl; firstorder.

    replace sm with sm2; auto.
    extensionality a; splitmaster.

    exists x; exists (fun a => match sm2 a with
                                 | None => x0 a
                                 | v => v
                               end); intuition eauto.

    unfold split, disjoint in *; intuition.
    erewrite H.
    eauto.
    rewrite H4.
    rewrite H6.
    eauto.

    splitmaster.

    eapply IHps1; eauto.
    unfold split, disjoint in *; intuition.
    eapply H.
    rewrite H4.
    specialize (H0 a0).
    destruct (x a0); eauto.
    specialize (H0 _ (refl_equal _)); congruence.

    splitmaster.
  Qed.

  Lemma sep'_ex : forall p k, sep' p k
    -> exists P, exists ps, k P ps
      /\ forall sm, P
        -> stars ps sm
        -> sep p sm.
    induction p; simpl; firstorder.

    specialize (IHp1 _ H); clear H; firstorder.
    specialize (IHp2 _ H); clear H; firstorder.
    do 2 eexists; split.
    eauto.
    rewrite stars_def; unfold sep; firstorder.
    apply stars'_app_fwd in H3; firstorder.
    rewrite stars_def in *.
    specialize (H1 _ H4 H7).
    specialize (H0 _ H2 H5).
    simpl.
    eapply sep'_weaken; [ eassumption | clear H0; firstorder ].
    eapply sep'_weaken; [ eassumption | clear H1; firstorder ].
    rewrite stars_def in *.
    eapply stars'_app_bwd; eauto.
    split; auto.

    specialize (H _ _ H0); clear H0; firstorder.
    do 2 eexists; split.
    eauto.
    intuition.
    specialize (H0 _ H1 H2).
    exists x.
    eauto.
  Qed.

  Lemma stars'_sep_bwd : forall ps2 sm,
    sep (fold_right Star (Pure True) (map Impure ps2)) sm
    -> stars' ps2 sm.
    induction ps2; simpl; firstorder.

    rewrite stars_def in *; firstorder.

    unfold sep in *; simpl in *.
    apply sep'_ex in H; firstorder.
    rewrite stars_def in *; firstorder.
    do 2 eexists; split.
    split; eauto.
    intuition.
    apply IHps2.
    specialize (H0 _ H2 H5).
    eapply sep'_weaken; [ eassumption | ].
    simpl; intuition.
    rewrite stars_def in *; assumption.
  Qed.
  
  Theorem desep_bwd : forall p ps2 ps1 sm,
    sep (fold_right Star (Pure True)
      (map Impure ps1 ++ p :: map Impure ps2)) sm
    -> stars (ps1 ++ sep p :: ps2) sm.
    induction ps1; simpl; firstorder.

    rewrite stars_def; firstorder.
    unfold sep in H; simpl in H.
    apply sep'_ex in H; firstorder.
    apply sep'_ex in H; firstorder.
    rewrite stars_def in *.
    apply stars'_app_fwd in H2; firstorder.
    do 2 eexists; split.
    split; eassumption.
    intuition.
    specialize (H1 _ H3 H6).
    apply stars'_sep_bwd; assumption.

    rewrite stars_def in *; unfold sep in *; firstorder.
    apply sep'_ex in H; firstorder.
    rewrite stars_def in *; firstorder.

    simpl.
    specialize (H0 _ H H1).
    unfold sep in H0; simpl in H0.
    apply sep'_ex in H0; firstorder.
    rewrite stars_def in *; firstorder.
    exists x3; exists x4; intuition.
    split; auto.
    apply IHps1.
    specialize (H2 _ H4 H7).
    eapply sep'_weaken.
    eassumption.
    simpl; intuition.
    rewrite stars_def in *; assumption.
  Qed.

  Theorem unfold_sep_bwd : forall p p' ps2 ps1 sm,
    sep (fold_right Star (Pure True)
      (map Impure ps1 ++ p :: map Impure ps2)) sm
    -> p ===> p'
    -> stars (ps1 ++ sep p' :: ps2) sm.
    intros; apply desep_bwd.

    generalize dependent sm.
    induction ps1; simpl; firstorder.
    unfold sep in *; simpl in *.
    red in H0.
    apply sep'_ex in H; firstorder.
    apply sep'_ex in H; firstorder.
    rewrite stars_def in *.
    apply stars'_app_fwd in H3; firstorder.
    specialize (H2 _ H4 H7).
    specialize (H1 _ H H5).
    apply sep'_weaken with (fun P ps => P /\ stars ps x3).
    apply H0; auto.
    intuition.
    eapply sep'_weaken; [ eauto | ].
    simpl; intuition.
    rewrite stars_def in *.
    eapply stars'_app_bwd; eauto.
    split; auto.

    unfold sep in H; simpl in H.
    apply sep'_ex in H; firstorder.
    rewrite stars_def in *; firstorder.
    specialize (H1 _ H3 H6).
    specialize (IHps1 _ H1).
    unfold sep; simpl.
    eapply sep'_weaken; [ eauto | ].
    rewrite stars_def in *; firstorder.
  Qed.

  Lemma stars_intro : forall p sm,
    sep p sm
    -> stars (sep p :: nil) sm.
    rewrite stars_def; firstorder.
    exists sm; exists (fun _ => None); intuition.
    splitmaster.
    constructor; auto.
  Qed.

  Lemma stars_elim : forall p sm,
    stars (sep p :: nil) sm
    -> sep p sm.
    rewrite stars_def; firstorder.
    replace sm with x; auto.
    extensionality a; splitmaster.
  Qed.

  Definition mem := M.addr -> M.byte.

  Definition mupd (m : mem) (a : M.addr) (v : M.byte) : mem := fun a' =>
    if M.addr_eq a' a then v else m a'.

  Theorem mupd_ne : forall m r1 v r2,
    r2 <> r1
    -> mupd m r1 v r2 = m r2.
    unfold mupd; intros; destruct (M.addr_eq r2 r1); intuition.
  Qed.

  Ltac simplifyMem' x a H' ls sm :=
    let Heq := fresh "Heq" in generalize H'; intro Heq;
      let rec finder pre post :=
        match eval hnf in post with
          | ptsTo ?a' ?v :: ?post =>
            apply (read (a := a') (a' := a) v post (rev pre) (sm := sm)) in Heq;
              [ | unfold M.addr; omega ]
          | ?p :: ?post => finder (p :: pre) post
        end in
        finder (@nil hprop) ls; injection Heq; clear Heq; intro Heq; rewrite Heq in *; clear Heq.

  Ltac doMupd m a v a' :=
    match goal with
      | [ _ : stars ?ls _ |- _ ] =>
        (let rec walk pre post :=
          match eval hnf in post with
            | ptsTo _ _ :: ?post =>
              let rec walk' pre' post :=
                match eval hnf in post with
                  | ptsTo _ _ :: ?post =>
                    let finishIt := eapply (ptrsNeq (rev pre) (rev pre') post); simpl; [
                      eassumption | solve [ auto ] | solve [ auto ] ] in
                    finishIt || (apply sym_not_eq; finishIt)
                  | ?p :: ?post =>
                    walk' (p :: pre') post
                end in
                walk' (@nil hprop) post
            | ?p :: ?post => walk (p :: pre) post
          end in
          let H := fresh "H" in assert (H : a' <> a); [ walk (@nil hprop) ls
            | rewrite (mupd_ne m (r1 := a) v (r2 := a') H) in *; clear H ]) || fail 1
    end.

  Ltac simplifyMem :=
    match goal with
      | [ H : stars ?ls ?sm |- context[M.Mem ?x ?a ] ] => simplifyMem' x a H ls sm
      | [ _ : context[M.Mem ?x ?a], H : stars ?ls ?sm |- _ ] => simplifyMem' x a H ls sm
      | [ |- context[mupd ?m ?a ?v ?a'] ] => doMupd m a v a'
      | [ _ : context[mupd ?m ?a ?v ?a'] |- _ ] => doMupd m a v a'
    end.

  Ltac ensureStars := match goal with
                        | [ |- stars _ _ ] => idtac
                      end.

  Ltac propSimpFwd := try tauto;
    repeat match goal with
             | [ H : True |- _ ] => clear H
             | [ H : ex _ |- _ ] => destruct H
             | [ H : _ /\ _ |- _ ] => destruct H
             | [ |- _ /\ _ ] => split
           end; try (tauto || (ensureStars; intuition)).

  Ltac propSimp := try tauto;
    repeat match goal with
             | [ H : True |- _ ] => clear H
             | [ H : ex _ |- _ ] => destruct H
             | [ H : _ /\ _ |- _ ] => destruct H
             | [ |- ex _ ] => esplit
             | [ |- _ /\ _ ] => split
           end; try (tauto || (ensureStars; intuition)).

  Ltac sepSimp'' := simpl in *; propSimpFwd; try subst.

  Ltac sepSimp' := sepSimp'';
    repeat match goal with
             | [ |- sep _ _ ] => hnf
             | [ H : sep _ _ |- _ ] => hnf in H
           end; sepSimp''.

  Ltac sepSimp := sepSimp'; repeat simplifyMem.

  Ltac sepSimpFull'' := simpl in *; propSimp; try subst.

  Ltac sepSimpFull' := sepSimpFull'';
    repeat match goal with
             | [ |- sep _ _ ] => hnf
             | [ H : sep _ _ |- _ ] => hnf in H
           end; sepSimpFull''.

  Inductive hide (P : Prop) := Hide : P -> hide P.
  Theorem hide_inv : forall P, hide P -> P.
    inversion 1; auto.
  Qed.

  Ltac propxFoHyps := apply hide_inv; propxFo; constructor.

  Ltac sepSimpFull := propxFoHyps; sepSimpFull'; repeat simplifyMem.

  Ltac memOut E :=
    match E with
      | (fun a' => Some (?f ?m ?a ?v a')) => constr:(f m a v)
      | (fun a' => ?f ?m ?a ?v a') => constr:(f m a v)
    end.

  Ltac mupdOut E :=
    match E with
      | _ ?m ?a ?v => constr:(m, a, v)
    end.

  Ltac equate x y := let H := fresh "H" in assert (H : x = y) by reflexivity; clear H.
  Ltac ensureEq x y := let H := fresh "H" in assert (H : x = y) by (unfold M.addr, M.byte in *; omega); clear H.

  Ltac reallyUnfold f := unfold f; try (let E := eval unfold f in f in unfold E).

  Ltac memsEq := intros;
    try match goal with
          | [ |- _ = Some (?f _ _ _ _) ] => reallyUnfold f
        end;
    try (let a := fresh "a" in extensionality a);
      unfold mupd, M.addr_eq; trivial; intros; simpl;
        repeat match goal with
                 | [ |- context[if ?E then _ else _] ] => destruct E; try reflexivity
               end; reflexivity.

  Ltac matchPtsTo :=
    repeat match goal with
             | [ _ : stars ?ls1 ?m1 |- stars ?ls2 ?m2 ] =>
               let H := fresh "H" in assert (H : m1 = m2) by memsEq; clear H;
                 match ls2 with
                   | context[ptsTo ?a ?v] =>
                     match v with
                       | v => fail 1
                       | _ => match ls1 with
                                | context [ptsTo ?a' ?v'] =>
                                  match v' with
                                    | v => fail 1
                                    | _ => ensureEq a a'; equate v v'
                                  end
                              end
                     end
                 end
           end.

  Module Type UNFOLD.
    Parameter should_unfold : sprop -> Prop.

    Axiom should_unfold_any : forall p, should_unfold p.
  End UNFOLD.

  Module Unfold : UNFOLD.
    Definition should_unfold (_ : sprop) := True.

    Theorem should_unfold_any : forall p, should_unfold p.
      red; auto.
    Qed.
  End Unfold.

  Import Unfold.
  Export Unfold.

  Lemma should_unfold_emp : should_unfold emp.
    intros; apply should_unfold_any.
  Qed.

  Lemma should_unfold_pure : forall P, should_unfold (Pure P).
    intros; apply should_unfold_any.
  Qed.

  Lemma should_unfold_star : forall p1 p2, should_unfold (p1 * p2).
    intros; apply should_unfold_any.
  Qed.

  Lemma should_unfold_ex : forall T (p1 : T -> _), should_unfold (Exis p1).
    intros; apply should_unfold_any.
  Qed.

  Hint Immediate should_unfold_emp should_unfold_pure should_unfold_star should_unfold_ex : Forward Backward.

  Ltac unfolderF :=
    match goal with
      | [ H : stars ?ls _ |- _ ] =>
        let rec loop pre post :=
          match eval hnf in post with
            | ?p :: ?post =>
              match p with
                | sep ?p' =>
                  (match p' with
                     | Pure ?P => apply (stars_elim_Pure P post (rev pre)) in H
                   end)
                  || (let H' := fresh "H" in assert (H' : should_unfold p'); [ solve [ eauto with Forward ] |
                    clear H'; apply (desep_fwd p' post (rev pre)) in H ])
                  || match goal with
                       | [ H' : ?P |- _ ] =>
                         match type of P with
                           | Prop => eapply (unfold_sep_fwd (p := p') post (rev pre)) in H; [ | eapply H'; solve [ eauto ] ]; clear H'
                         end
                     end
                  || (eapply (unfold_sep_fwd (p := p') post (rev pre)) in H; [ | solve [ eauto with Forward ] ]);
                  sepSimp
                | _ => loop (p :: pre) post
              end
          end in
          loop (@nil hprop) ls
    end.

  Ltac updater E :=
    try (let p := mupdOut E in
      match p with
        | (?m, ?a, ?v) => updater m;
          match goal with
            | [ H : stars ?ls _ |- _ ] =>
              let rec overwriter pre post :=
                match eval hnf in post with
                  | nil => idtac
                  | ptsTo ?a' ?v' :: ?post => 
                    apply (overwrite (a := a') (a' := a)
                      v' v post (rev pre)) in H;
                    [ | solve [ auto ] ]
                  | ?p :: ?post => overwriter (p :: pre) post
                end in
                overwriter (@nil hprop) ls
          end
      end).

  Ltac sepUpdate :=
    match goal with
      | [ |- stars _ ?E ] => let m := memOut E in updater m; simpl in *; matchPtsTo
      | _ => matchPtsTo
    end.

  Ltac unfolderB :=
    match goal with
      | [ |- stars ?ls _ ] =>
        let rec loop pre post :=
          match eval hnf in post with
            | ?p :: ?post =>
              match p with
                | sep ?p' =>
                  (let H := fresh "H" in assert (H : should_unfold p'); [ solve [ auto with Backward ] |
                    clear H; apply (desep_bwd p' post (rev pre)) ])
                  || match goal with
                       | [ H : ?P |- _ ] =>
                         match type of P with
                           | Prop => eapply (unfold_sep_bwd (p' := p') post (rev pre)); [ clear H |
                             eapply H; clear H; solve [ eauto ] ]
                         end
                     end
                  || (eapply (unfold_sep_bwd (p' := p') post (rev pre)); [ |
                    solve [ eauto with Backward ] ]);
                  sepSimpFull
                | _ => loop (p :: pre) post
              end
          end in
          loop (@nil hprop) ls
    end.

  Ltac sepFwd1 := unfolderF; try autorewrite with InSep in *.
  Ltac sepFwd := repeat sepFwd1.

  Ltac sepBwd1 := unfolderB; repeat simplifyMem; matchPtsTo; try autorewrite with InSep in *.
  Ltac sepBwd := repeat sepBwd1.

  Theorem sep_cong : forall x y, x = y -> sep x = sep y.
    congruence.
  Qed.

  Ltac eqer := unfold M.addr; solve [ auto; repeat ((apply sep_cong || (progress f_equal)); auto) ].

  Ltac canceler :=
    repeat match goal with
             | _ => apply himp_refl
             | _ => apply frame_finish
             | [ |- ?ps1 ---> ?ps2 ] =>
               let rec loop1 pre1 post1 :=
                 match eval hnf in post1 with
                   | ?p1 :: ?post1 =>
                     let rec loop2 pre2 post2 :=
                       match eval hnf in post2 with
                         | ?p2 :: ?post2 =>
                           (ensureNotUnif p2; let Ht := fresh "Ht" in
                             generalize (himp_cancel_eq (rev pre1) post1 (rev pre2) post2); intro Ht; simpl in Ht;
                               apply Ht; clear Ht; [ eqer | ])
                           || loop2 (p2 :: pre2) post2
                       end in
                       loop2 (@nil hprop) ps2
                       || loop1 (p1 :: pre1) post1
                 end in
                 loop1 (@nil hprop) ps1
           end.

  Ltac sepCancel := try apply stars_elim;
    try match goal with
          | [ H : sep _ _ |- _ ] => apply stars_intro in H
          | [ H : sep' _ _ |- _ ] => apply stars_intro in H
        end;
    try match goal with
          | [ H : stars _ _ |- _ ] => apply (stars_final _ H); [ memsEq | clear H; canceler ]
        end.

  Ltac sepLemma := unfold himp, Himp; intuition; sepSimp; auto; sepFwd; sepSimpFull; sepBwd; sepCancel.
  Ltac sep := sepSimp; try autorewrite with InSep in *; sepFwd; sepSimpFull;
    sepUpdate; try autorewrite with InSep in *; sepBwd; sepCancel.

  Hint Extern 2 (@eq nat _ _) => omega.
  Hint Extern 2 (not (@eq nat _ _)) => omega.

  Definition sempty : smem := fun _ => None.
End Make.
